/******************************************************************************
 * Deck    : Pioneer DDJ-SB
 * Author  : Erik Häll
 * Version : 0.3
*/

var PioneerDDJSB = function() { }

PioneerDDJSB.debugging = true;
PioneerDDJSB.LEDinterval = 90;
PioneerDDJSB.beatLEDinterval = 20;
PioneerDDJSB.statusReadInterval = 250;
PioneerDDJSB.blinkState = 0;
PioneerDDJSB.pausePattern = [1,1,1,1,1,0,0,0];
PioneerDDJSB.scratchPattern = [0,1,0,1,0,1,0,1];
PioneerDDJSB.isGoingToEndLEDstate  = [1,1,1,1,0,0,0,0];
PioneerDDJSB.isGoingToEndLEDstate2 = [1,1,0,0,1,1,0,0];

PioneerDDJSB.cueLEDstate = [0,0];
PioneerDDJSB.playLEDstate = [0,0];
PioneerDDJSB.syncLEDstate = [0,0];

PioneerDDJSB.durationStatus = [0.0,0.0];
PioneerDDJSB.playStatus = [0,0];
PioneerDDJSB.playPosition = [0.0,0.0];

PioneerDDJSB.debug = function(message) 
{
    if( PioneerDDJSB.debugging )
    	print("Pioneer DDJ-SB: " + message);
}

PioneerDDJSB.init = function(id)
{
	var ALPHA = 1.0 / 8;
	var BETA = ALPHA / 32;
	var JOG_RESOLUTION = 720;
	var VINYL_SPEED = 33 + 1/3;
	var NUMBER_OF_ACTIVE_PERFORMANCE_PADS = 4;
	var SAFE_SCRATCH_TIMEOUT = 20; // 20ms is the minimum allowed here.
	
	PioneerDDJSB.settings = 
	{
		alpha: ALPHA,
		beta: BETA,
		jogResolution: JOG_RESOLUTION,
		vinylSpeed: VINYL_SPEED,
		numberOfActivePerformancePads: NUMBER_OF_ACTIVE_PERFORMANCE_PADS,
		safeScratchTimeout: SAFE_SCRATCH_TIMEOUT
	};
		
		
	PioneerDDJSB.channels = 
	{	
		0x00: {},
		0x01: {}
	};
		
	PioneerDDJSB.channelGroups =
	{
		'[Channel1]': 0x00,
		'[Channel2]': 0x01
	};
	
	PioneerDDJSB.loopIntervals =
	{
		PAD1: 1, 
		PAD2: 2, 
		PAD3: 4, 
		PAD4: 8
	};
	
	PioneerDDJSB.rotarySelectorTargets =
	{
		libraries: 0,
		tracklist: 1
	},
	
	// Initialize the Decks and the Library
	PioneerDDJSB.rotarySelectorMode = 1;
				
	PioneerDDJSB.BindControlConnections(false);

	//set LED timer
	PioneerDDJSB.ledTimer = engine.beginTimer(PioneerDDJSB.LEDinterval, "PioneerDDJSB.setLeds()");
	PioneerDDJSB.beatLedTimer = engine.beginTimer(PioneerDDJSB.beatLEDinterval, "PioneerDDJSB.setBeatLeds()");
	PioneerDDJSB.statusReadTimer = engine.beginTimer(PioneerDDJSB.statusReadInterval, "PioneerDDJSB.statusRead()");


	// Setup mixer 
	// TO-DO

	// Done init
	PioneerDDJSB.debug("Controller Initialized");
};


PioneerDDJSB.resetLEDTimer = function () {
	engine.stopTimer(PioneerDDJSB.ledTimer);
	engine.stopTimer(PioneerDDJSB.beatLedTimer);
	engine.stopTimer(PioneerDDJSB.statusReadTimer);
	PioneerDDJSB.setLeds()
	PioneerDDJSB.ledTimer = engine.beginTimer(PioneerDDJSB.LEDinterval, "PioneerDDJSB.setLeds()");
};


PioneerDDJSB.elapsedTime = function (deck) 
{
    return engine.getValue("[Channel"+deck+"]","duration") * engine.getValue("[Channel"+deck+"]","playposition");
}


// Only for timer read of status
PioneerDDJSB.statusRead = function () 
{
	for (var channelIndex = 1; channelIndex <= 2; channelIndex++)
	{
		var channelGroup = '[Channel' + channelIndex + ']';
		PioneerDDJSB.durationStatus[channelIndex] = engine.getValue(channelGroup, "duration");
		PioneerDDJSB.playPosition[channelIndex]   = engine.getValue(channelGroup, "playposition");
		PioneerDDJSB.playStatus[channelIndex]     = engine.getValue(channelGroup, 'play');
	}
}


PioneerDDJSB.isGoingToEnd = function (duration,playposition) 
{
	var secondsBlink = 30.0;
    	var secondsToEnd = duration * (1-playposition);

	if (secondsToEnd < secondsBlink && 
	    secondsToEnd > 1 ) 
	{ 
		// The song is going to end
		return secondsToEnd;
	}
	return 0;
}


PioneerDDJSB.setLeds = function () {
	// Stop timer to avoid timing problem
	engine.stopTimer(PioneerDDJSB.ledTimer);

	for (var channelIndex = 1; channelIndex <= 2; channelIndex++)
	{
		var channelGroup = '[Channel' + channelIndex + ']';
		var cueState=0;
		var playState=0;
		var playposition = PioneerDDJSB.playPosition[channelIndex];
		var duration = PioneerDDJSB.durationStatus[channelIndex];

		if( duration == 0) 
		{
			if( PioneerDDJSB.blinkState==0 )
			{
				playState=1; // Play / Pause LED on
				cueState=1; // Cue LED on
			}else{
				playState=0; // Play / Pause LED off
				cueState=0; // Cue LED off
			}
	    	}else{
			var isScratching = engine.isScratching(channelIndex);
			if ( isScratching )
			{
				if( PioneerDDJSB.pausePattern[PioneerDDJSB.blinkState] )
					playState=1; // Play / Pause LED on
				else
					playState=0; // Play / Pause LED off
			}

			if( PioneerDDJSB.playStatus[channelIndex] )
			{
				goingToEnd = PioneerDDJSB.isGoingToEnd( duration, playposition );
				if( goingToEnd != 0 )
				{
					// Cue LED indicating end of playback
					if( goingToEnd > 20.0 )
						cueState = PioneerDDJSB.isGoingToEndLEDstate[PioneerDDJSB.blinkState];  
					else
						cueState = PioneerDDJSB.isGoingToEndLEDstate2[PioneerDDJSB.blinkState];
				}else{
					cueState=0; // Cue LED off
				}

				if( !isScratching )
					if( goingToEnd < 20.0 &&  goingToEnd > 0)
					{
						playState = PioneerDDJSB.isGoingToEndLEDstate2[PioneerDDJSB.blinkState]^1;
					}else{
						playState=1; // Play / Pause LED on
					}
			}else
			if( PioneerDDJSB.pausePattern[PioneerDDJSB.blinkState]==0 )
				cueState=0; // Cue LED off
			else
				cueState=1; // Cue LED on
		}

		if( PioneerDDJSB.playLEDstate[channelIndex] != playState )
		{
			PioneerDDJSB.playLEDstate[channelIndex] = playState;
			midi.sendShortMsg(0x90+channelIndex-1, 0x0B, playState ? 0x7F : 0x00); // play/pause LED
		}
		if( PioneerDDJSB.cueLEDstate[channelIndex] != cueState )
		{
			PioneerDDJSB.cueLEDstate[channelIndex] = cueState;
			midi.sendShortMsg(0x90+channelIndex-1, 0x0C, cueState ? 0x7F : 0x00); // Cue LED
		}
	}

	PioneerDDJSB.blinkState=PioneerDDJSB.blinkState+1;
	if( PioneerDDJSB.blinkState >= 8 )
		PioneerDDJSB.blinkState=0;

	// Turn on timer again
	PioneerDDJSB.ledTimer = engine.beginTimer(PioneerDDJSB.LEDinterval, "PioneerDDJSB.setLeds()");
};


PioneerDDJSB.setBeatLeds = function () {

	for (var channelIndex = 1; channelIndex <= 2; channelIndex++)
	{
		var channelGroup = '[Channel' + channelIndex + ']';

		var beatSyncValue = 0;

		// Invers LED when beatsync is active
		if( engine.getValue(channelGroup, "beatsync") )
			beatSyncValue = 1;

		//engine.getValue(channelGroup, "beatsync"); 
		//engine.getValue(channelGroup, "quantize");
	
		if (engine.getValue(channelGroup, "play") &&
		    engine.getValue(channelGroup, "beat_active")) 
		{
			beatSyncValue = !beatSyncValue;
		}

		if( PioneerDDJSB.syncLEDstate[channelIndex] != beatSyncValue )
		{
			PioneerDDJSB.syncLEDstate[channelIndex] = beatSyncValue;

			if( beatSyncValue )
			{
				midi.sendShortMsg(0x90+channelIndex-1, 0x58, 0x7F); // Sync LED	
			}else{
				midi.sendShortMsg(0x90+channelIndex-1, 0x58, 0x00); // Sync LED	
			}
		}
	}
	/**/

};


PioneerDDJSB.shutdown = function()
{
	engine.stopTimer(PioneerDDJSB.ledTimer);
	engine.stopTimer(PioneerDDJSB.beatLedTimer);
	PioneerDDJSB.BindControlConnections(true);
};

PioneerDDJSB.BindControlConnections = function(isUnbinding)
{
	for (var channelIndex = 1; channelIndex <= 2; channelIndex++)
	{
		var channelGroup = '[Channel' + channelIndex + ']';
		
		// Play / Pause LED
		engine.connectControl(channelGroup, 'play', 'PioneerDDJSB.PlayLeds', isUnbinding);
		
		// Cue LED
		engine.connectControl(channelGroup, 'cue_default', 'PioneerDDJSB.CueLeds', isUnbinding);
		
		// PFL / Headphone Cue LED
		engine.connectControl(channelGroup, 'pfl', 'PioneerDDJSB.HeadphoneCueLed', isUnbinding);
		
		// Keylock LED
		engine.connectControl(channelGroup, 'keylock', 'PioneerDDJSB.KeyLockLeds', isUnbinding);
		
		// Vinyl LED
		engine.connectControl(channelGroup, 'slip_enabled', 'PioneerDDJSB.ToggleVinylLed', isUnbinding);
		
		// Hook up the hot cue performance pads
		for (var i = 0; i < 4; i++)
		{
			engine.connectControl(channelGroup, 'hotcue_' + (i + 1) +'_enabled', 'PioneerDDJSB.HotCuePerformancePadLed', isUnbinding);
		}
		
		// Hook up the roll performance pads
		for (var i = 1; i <= PioneerDDJSB.settings['numberOfActivePerformancePads']; i++)
		{
			engine.connectControl(channelGroup, 'beatloop_' + PioneerDDJSB.loopIntervals['PAD' + i] + '_enabled', 'PioneerDDJSB.RollPerformancePadLed', isUnbinding);
		}
	}
};


///////////////////////////////////////////////////////////////
//                         MIX SECTION                       //
///////////////////////////////////////////////////////////////

/*PioneerDDJSB.headMix = function (channel, control, value, status, group) 
{
	PioneerDDJSB.debug("headmix " + value);

	engine.setValue("[Master]","headMix",(value-64)/63);
};*/

///////////////////////////////////////////////////////////////
//                         LED SECTION                       //
///////////////////////////////////////////////////////////////

// This handles LEDs related to the PFL / Headphone Cue event.
PioneerDDJSB.HeadphoneCueLed = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];	
	midi.sendShortMsg(0x90 + channel, 0x54, value ? 0x7F : 0x00); // Headphone Cue LED
};

// This handles LEDs related to the cue_default event.
PioneerDDJSB.CueLeds = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];	
	midi.sendShortMsg(0x90 + channel, 0x0C, value ? 0x7F : 0x00); // Cue LED
	PioneerDDJSB.cueLEDstate[channel] = value;
};

// This handles LEDs related to the keylock event.
PioneerDDJSB.KeyLockLeds = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];	
	midi.sendShortMsg(0x90 + channel, 0x1A, value ? 0x7F : 0x00); // Keylock LED
};

// This handles LEDs related to the play event.
PioneerDDJSB.PlayLeds = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];	
	midi.sendShortMsg(0x90 + channel, 0x0B, value ? 0x7F : 0x00); // Play / Pause LED
	midi.sendShortMsg(0x90 + channel, 0x0C, value ? 0x7F : 0x00); // Cue LED
	PioneerDDJSB.playLEDstate[channel] = value;
	PioneerDDJSB.cueLEDstate[channel] = value;
};

PioneerDDJSB.ToggleVinylLed = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];
	midi.sendShortMsg(0x90 + channel, 0x17, value ? 0x7F : 0x00); // Vinyl LED
};


// Lights up the LEDs for beat-loops.
PioneerDDJSB.RollPerformancePadLed = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];
	for (var i = 1; i <= PioneerDDJSB.settings['numberOfActivePerformancePads']; i++)
	{
		if (control === 'beatloop_' + PioneerDDJSB.loopIntervals['PAD' + i] + '_enabled')
		{
			midi.sendShortMsg(0x97 + channel, 0x10 + (i - 1), value ? 0x7F : 0x00); 
		}
	}
};

// Lights up the LEDs for the hotcue buttons. 
// The buttons LEDs are disabled using the SHIFT button
PioneerDDJSB.HotCuePerformancePadLed = function(value, group, control) 
{
	var channel = PioneerDDJSB.channelGroups[group];
	
	var padIndex = null;
	for (var i = 0; i < 8; i++)
	{
		if (control == 'hotcue_' + i + '_enabled')
		{
			break;
		}
		
		padIndex = i;
	}
	
	// Pad LED without shift key
	midi.sendShortMsg(0x97 + channel, 0x00 + padIndex, value ? 0x7F : 0x00);
	
	// Pad LED with shift key
	midi.sendShortMsg(0x97 + channel, 0x00 + padIndex + 0x08, value ? 0x7F : 0x00);
};

///////////////////////////////////////////////////////////////
//                     JOGWHEEL SECTION                      //
///////////////////////////////////////////////////////////////

// Work out the jog-wheel change / delta
// The Wheel control centers on 0x40.
PioneerDDJSB.getJogWheelDelta = function(value)
{
	return value - 0x40;  	// find out how much it's moved.
};

// Toggle scratching for a channel
PioneerDDJSB.toggleScratch = function(channel, isEnabled)
{
	PioneerDDJSB.debug("toggleScratch " + channel + " " + isEnabled);

	var deck = channel + 1; 
	if (isEnabled) 
	{
		engine.scratchEnable(
			deck, 
		PioneerDDJSB.settings['jogResolution'], 
		PioneerDDJSB.settings['vinylSpeed'], 
		PioneerDDJSB.settings['alpha'], 
		PioneerDDJSB.settings['beta']);
	}else{
		engine.scratchDisable(deck);
	}
};

// Detect when the user touches and releases the jog-wheel while 
// jog-mode is set to vinyl to enable and disable scratching.
PioneerDDJSB.jogScratchTouch = function(channel, control, value, status, group) 
{
	PioneerDDJSB.debug("jogScratchTouch " + channel + " " + value + " " + group);
	
	var deck = channel + 1; 
	
	if (!engine.getValue(group, 'play'))
	{
		PioneerDDJSB.toggleScratch(channel, value == 0x7F);
	}
	else
	{
		var activate = value > 0;
		
		if (activate) 
		{
			engine.brake(deck, true, 1, 1); // enable brake effect
			PioneerDDJSB.toggleScratch(channel, true);
		}
		else 
		{
			engine.brake(deck, false, 1, 1); // disable brake effect
			PioneerDDJSB.toggleScratch(channel, false);
		}  
	}
};
 
// Scratch or seek with the jog-wheel.
PioneerDDJSB.jogScratchTurn = function(channel, control, value, status) 
{
	var deck = channel + 1; 
	
	// Only scratch if we are in scratching mode.
        // (touching the top of the jog-wheel)
	if (engine.isScratching(deck)) 
	{
		engine.scratchTick(deck, PioneerDDJSB.getJogWheelDelta(value));
	}
};

// Pitch bend using the jog-wheel or finish a scratch when the wheel 
// is still turning after releasing it.
PioneerDDJSB.jogPitchBend = function(channel, control, value, status, group) 
{
	PioneerDDJSB.debug("jogPitchBend " + channel + " " + value + " " + group);
	var deck = channel + 1; 

	if ( engine.getValue(group, 'play') )
	{
		PioneerDDJSB.pitchBend(channel, PioneerDDJSB.getJogWheelDelta(value));
	}else{
	}
};

// Pitch bend a channel
PioneerDDJSB.pitchBend = function(channel, movement) 
{
	PioneerDDJSB.debug("pitchBend " + channel + " " + movement);

	var deck = channel + 1; 
	var group = '[Channel' + deck +']';
	
	// Make this a little less sensitive.
	movement = movement / 5; 
	
	// Limit movement to the range of -3 to 3.
	movement = movement > 3 ? 3 : movement;
	movement = movement < -3 ? -3 : movement;
	
	engine.setValue(group, 'jog', movement);
};

// Called when the jog-mode is not set to vinyl, and the jog wheel is touched.
// If we are not playing we want to seek through it and this is done in scratch mode
PioneerDDJSB.jogSeekTouch = function(channel, control, value, status, group) 
{
	PioneerDDJSB.debug("jogSeekTouch " + channel + " " + value + " " + group);
	if( !engine.getValue(group, 'play') )
	{
		//var deck = channel + 1; 
		PioneerDDJSB.toggleScratch(channel, value == 0x7F);
	}
};

// Call when the jog-wheel is turned. The related jogSeekTouch function 
// sets up whether we will be scratching or pitch-bending depending 
// on whether a song is playing or not.
PioneerDDJSB.jogSeekTurn = function(channel, control, value, status, group) 
{
	PioneerDDJSB.debug("jogSeekTurn " + channel + " " + value + " " + group);

	if( !engine.getValue(group, 'play') )
	{
		var deck = channel + 1; 
		engine.scratchTick(deck, PioneerDDJSB.getJogWheelDelta(value));
	}
};

///////////////////////////////////////////////////////////////
//                     SELECTOR SECTION                      //
///////////////////////////////////////////////////////////////

/**********************************************************************
* This handles the rotary selector for:
*  - choosing tracks 
*  - library items
*  - crates
***********************************************************************/
PioneerDDJSB.RotarySelector = function(channel, control, value, status) 
{
	var delta = 0x40 - Math.abs(0x40 - value);
	var isCounterClockwise = value > 0x40;
	if (isCounterClockwise)
	{
		delta *= -1;
	}
	
	var tracklist = PioneerDDJSB.rotarySelectorTargets.tracklist;
	var libraries = PioneerDDJSB.rotarySelectorTargets.libraries;
	
	switch(PioneerDDJSB.rotarySelectorMode)
	{
		case tracklist:
			engine.setValue('[Playlist]', 'SelectTrackKnob', delta);
			break;
		case libraries:
			if (delta > 0)
			{
				engine.setValue('[Playlist]', 'SelectNextPlaylist', 1);
			}
			else if (delta < 0)
			{
				engine.setValue('[Playlist]', 'SelectPrevPlaylist', 1);
			}
			
			break;
	}
};



PioneerDDJSB.RotarySelectorClick = function(channel, control, value, status) 
{
	/*********************************************************************
	*  Trig when the button is pressed down.
	*********************************************************************/
	if( value == 0x7F )
	{
		if( PioneerDDJSB.rotarySelectorMode == 0 ) // library
		{
			if( engine.getValue('[Playlist]', 'ToggleSelectedSidebarItem') == 1 )
				engine.setValue('[Playlist]', 'ToggleSelectedSidebarItem', 0);
			else
				engine.setValue('[Playlist]', 'ToggleSelectedSidebarItem', 1);
		}
	}
};

PioneerDDJSB.backClick = function(channel, control, value, status)
{
	/*********************************************************************
	*  Trig when the button is pressed down.
	*********************************************************************/
	if (value == 0x7F)
	{
		if( PioneerDDJSB.rotarySelectorMode == 1 )
			PioneerDDJSB.rotarySelectorMode = 0;
		else
			PioneerDDJSB.rotarySelectorMode = 1;
	}
};

